unit crtmenu;
INTERFACE

  { Author S. Kleiner. Released to Public Domain. }

  { CRTMENU UNIT.

    Contains cmenu function.


    Used to display "Bar Menus" on the crt and get an answer from the
    keyboard.  The user may make a selection by several means:

    1.  Use the movement keys to highlight a menu selection
        (Left-Arrow [also Space, Tab], Right-Arrow [also Backspace],
         Home, and End)

    2.  Use the function keys (F1 .. F9) to highlight a menu selection

    3.  Use the first character of a menu selection to highlight it.

    After highlighting the selection, the user presses the RETURN key
    (designated by CM_RET_KY) to accept it.  The user can also select to
    Escape the menu by pressing the ESCAPE key (designated by CM_ESC_KY).

    The function returns the selection number from 1 .. CM_MAX_SEL or
    zero if the ESCAPE key was entered.

    If CM_AUTOCLEAR is true (default) or the user enters the ESCAPE key,
    the menu is cleared from the screen on exit, otherwise is is left on
    the screen.

    -----------------------------------------------------------------------

    Using the cmenu function:


    MENULINE Parameter:

           The menuline parameter indicates the line on the CRT to use
           to display the menu.  The first line is zero (0).  The last
           line is 23.


    MENUSTR Parameter:

           The menu is entered as a string in the following format:

           'selection1|selection2| ... selection9|'

    Note:  There are a maximum of CM_MAX_SEL = 9 selections allowed.
           Each selection may contain no blanks and must be followed
           by the BAR (|) character.  Blanks may be embedded in the
           menu selections by using the TILDE character (~) which will
           be replaced in the menu display by a blank.  The menu may be
           at most CM_MAX_LEN characters long.

    Use of Automatic Function Key Labels:

           If the first three characters of a Menu Selection are:
           '\\:', then the backslash characters will automatically
           be replaced with 'Fx' where x is the corresponding selection
           number.  E.g.:

           '\\:Add|\\:Delete|\\:Change|\\:Next|\\:Previous|'

           would display as:

           'F1:Add F2:Delete F3:Change F4:Next F5:Previous '

           But the user could still enter A,D,C,N, or P to make a
           selection by the first letter.


    DEFAULT Parameter

           The default parameter allows you to pre-select an option to
           be highlighted on the initial display of the menu.  To select
           the default, the user need only press RETURN.  If the default
           is less than 1 it is set to 1.  If it is greater than the
           last selection available, it is set to the last selection.

    CM_HelpProc Variable

           The CM_HelpProc variable is a procedure variable which allows
           for an explanation or description of the current selection to
           be displayed.  At initialization of the unit, this variable is
           set to the CM_No_Help procedure which does nothing.

           You may define your own procedure which displays help and is
           of the proper procedure heading format for CM_Help_Proc
           procedures.  Your procedure must be declared FAR using the
           $F+ compiler option.  To use your own set the CM_HelpProc
           variable by including a line like:

           CM_HelpProc := Your_Help_Proc;

           To go back to the default, use:

           CM_HelpProc := CM_No_Help;

           NOTE:  When called with helpsel = 0, your procedure should
           clear the information previously displayed.  If called with
           helpsel < 0, the final selection is abs(helpsel), but this
           indicates an "erase the help" call. Depending on your needs
           you may or may not want to clear the help. This is usefull
           for "pull-down" menus.

    CM_Highlight variable

           The CM_Highlight variable is a procedure variable which allows
           you to use different means to highlight the current menu
           selection.  When called with a "True" parameter value, it
           causes text written to the crt to be highlighted, when called
           with a "False" parameter value, it causes text written to the
           crt to be displayed normally.  At initialization, this variable
           is set to CM_HighIntensity, which uses the Crt Unit procedures
           HighVideo and NormVideo to set highlighting.  You may use your
           own procedure to show a different highlighting approach.
           It must be declared and installed similar to the CM_HelpProc
           above.

    CM_ErrProc variable

           As with the above two variables, this allows you to customize
           the error handling.  The standard procedure, CM_Error simply
           sounds the "bell" by writing a ^G to the console.  This
           procedure is passed the menuline as a parameter, but is not
           used in the standard procedure.

  }


const
  CM_ESC_KY   : char = #27; { ESC }
  CM_RET_KY   : char = ^M;  { ENTER }
  CM_AUTOCLEAR: boolean = true;
  CM_MAX_LEN  = 78;  { Max Len of Menu Line }
  CM_MAX_SEL  =  9;  { Max No. of choices on Menu Line }

type

  CM_Help_Proc = procedure (menuline : byte; { the line displaying the menu }
                            helpsel  : integer;
                                             { the current selection number,
                                               if = 0, then clear the help.
                                               if < 0, then abs(helpsel) is
                                               the final selection. }
                            selstrt  : byte; { the column where the current
                                               selection starts in the menu }
                            selend   : byte);{ the column where the current
                                               selection ends in the menu }

  CM_High_Intensity_Proc = procedure (hilite : boolean);

  CM_Error_Proc          = procedure (menuline : byte);


var

  CM_HelpProc : CM_Help_Proc; { called to show menu help, default is
                                do nothing.  When called with helpsel = 0,
                                should clear the help message }

  CM_Highlight: CM_High_Intensity_Proc; { called to turn highlighting
                                          on/off }

  CM_ErrProc  : CM_Error_Proc;

  function CM_Menu(menuline : byte; menustr : string; default : byte) : byte;

  procedure CM_No_Help ( menuline : byte;   { Default help procedure }
                         helpsel  : integer;{ Does nothing. }
                         selstrt  : byte;
                         selend   : byte);

  procedure CM_HighIntensity ( hilite : boolean );

  procedure CM_Error( menuline : byte );


IMPLEMENTATION
  uses crt, strnglib;

  procedure CM_No_Help (menuline : byte;
                        helpsel  : integer;
                        selstrt  : byte;
                        selend   : byte);
  begin { CM_No_Help }
    { do nothing }
  end; { CM_No_Help }

  procedure CM_HighIntensity ( hilite : boolean );
  begin { CM_HighIntensity }
    if hilite
      then HighVideo
      else NormVideo;
  end; { CM_HighIntensity }

  procedure CM_Error( menuline : byte );
  begin { CM_Error }
    write(^G);
  end; { CM_Error }

  function CM_Menu(menuline : byte; menustr : string; default : byte) : byte;
  const
    LT_ARW  = #128;
    RT_ARW  = #129;
    HOME_KY = #130;
    END_KY  = #131;
    F1_KY   = #132;
    F9_KY   = #140;
    CM_TAB  = ^I;
    CM_BS   = ^H;
    CM_SPC  = ' ';
    CM_ESC = #27;
    CM_RET = ^M;
  var
    sel   : byte;
    selindex : array [0 .. CM_MAX_SEL] of byte;
    selstart : array [1 .. CM_MAX_SEL] of char;
    ch    : char;
    mxsel : byte;
    fnd   : boolean;
    origsel : byte;

    procedure getextkey;
    const
      LT_IN_ARW  = #75;
      RT_IN_ARW  = #77;
      HOME_IN_KY = #71;
      END_IN_KY  = #79;
      F1_IN_KY   = #59;
      F9_IN_KY   = #67;
    begin { getextkey }
      ch := readkey;
      case ch of
        LT_IN_ARW : ch := LT_ARW;
        RT_IN_ARW : ch := RT_ARW;
        HOME_IN_KY: ch := HOME_KY;
        END_IN_KY : ch := END_KY;
        F1_IN_KY ..
        F9_IN_KY  : ch := chr(ord(F1_KY) + ord(ch) - ord(F1_IN_KY));
        ELSE ch := #0;
      end; { case }
    end; { getextkey }

  begin { CM_Menu }

    { Set Video Mode }
    CM_Highlight(false);

    { truncate menu string if too long }
    menustr := trim(menustr);
    { delete any blanks in menustr }
    for sel := 1 to length(menustr) do
      if menustr[sel] = ' '
        then delete(menustr,sel,1);
    if length(menustr) > CM_MAX_LEN
      then menustr := copy(menustr,1,CM_MAX_LEN);
    if menustr[length(menustr)] <> '|'
      then menustr := menustr + '|';

    { generate selection index }
    mxsel := 0;
    selindex[0] := 0;
    repeat
      sel := pos('|',menustr);
      if sel > 0
        then begin
          inc(mxsel);
          selindex[mxsel] := sel;
          selstart[mxsel] := Upcase(menustr[selindex[pred(mxsel)]+1]);
          if copy(menustr,selindex[pred(mxsel)]+1,3) = '\\:'
            then begin
              delete(menustr,selindex[pred(mxsel)]+1,2);
              insert('F'+chr(ord('0') + mxsel),menustr,
                     selindex[pred(mxsel)]+1);
              selstart[mxsel] := menustr[selindex[pred(mxsel)]+4];
            end;
          menustr[sel]    := ' ';
        end;
    until (sel = 0) or (mxsel >= CM_MAX_SEL);

    { convert Tilde to space }
    repeat
      sel := pos('~',menustr);
      if sel > 0
        then menustr[sel] := ' ';
    until (sel = 0);

    { display menu string }
    gotoxy(1,menuline);
    clreol;
    write(menustr);

    { clear the help line in case background is different }
    CM_HelpProc(menuline,0,0,0);

    { set to default }
    sel := default;
    if sel < 1
      then sel := 1
      else if sel > mxsel
        then sel := mxsel;

    { get the selection }
    repeat
      { highlight current selection }
      gotoxy(selindex[pred(sel)]+1,menuline);
      CM_Highlight(true);
      write(copy(menustr,
                 succ(selindex[pred(sel)]),
                 pred(selindex[sel])-selindex[pred(sel)]
                )
           );
      CM_Highlight(false);

      { Show help information if any }
      CM_HelpProc (menuline,sel,selindex[pred(sel)]+1,
                                 selindex[sel]-1);

      { get a char/key }
      ch := Upcase(readkey);
      if ch = #0
        then getextkey
        else if (ch=CM_ESC_KY)
          then ch := CM_ESC
          else if (ch=CM_RET_KY)
            then ch := CM_RET;

      { "UN-Highlight the current selection" }
      gotoxy(selindex[pred(sel)]+1,menuline);
      write(copy(menustr,
                  succ(selindex[pred(sel)]), { starts at pred + 1 }
                  pred(selindex[sel])-selindex[pred(sel)]
                )
           );

      { handle the key }
      case ch of
        CM_BS,
        LT_ARW : if sel > 1
                   then dec(sel)
                   else sel := mxsel;
        CM_TAB,
        CM_SPC,
        RT_ARW : if sel < mxsel
                   then inc(sel)
                   else sel := 1;
        HOME_KY: sel := 1;
        END_KY : sel := mxsel;

        CM_ESC : sel := 0;
        CM_RET : { do nothing, will exit next };
        F1_KY..
        F9_KY  : begin
                   origsel := sel;
                   sel := ord(ch) - ord(F1_KY) + 1;
                   if sel > mxsel
                     then begin
                       CM_ErrProc(menuline);
                       sel := origsel;
                     end;
                 end;
        ELSE begin
          { find next selstart with ch = selstart, if none, then err bell }
          origsel := sel; { remember where we started }
          fnd     := false;
          repeat
            inc(sel);
            if (sel > mxsel)
              then sel := 1;
            fnd := (selstart[sel] = ch);
          until fnd or (sel = origsel);
          if not fnd
            then begin
              CM_ErrProc(menuline);
              sel := origsel;
            end;
        end; { ELSE }
      end; { case }
      if (ch = CM_RET)
        then CM_HelpProc(menuline,-sel,selindex[pred(sel)]+1,
                                       selindex[sel]-1) { clear or keep }
        else CM_HelpProc(menuline,0,0,0); { clear the help }
    until (ch = CM_RET) or (ch = CM_ESC);
    if (ch = CM_ESC) or (CM_AUTOCLEAR)
      then begin
        gotoxy(1,menuline);
        clreol;
      end;
    CM_Menu := sel;
  end; { CM_Menu }

begin { CRTMENU }

  CM_HelpProc := CM_No_Help;  { default is no help line }
  CM_Highlight:= CM_HighIntensity; { default is HighVideo/NormVideo }
  CM_ErrProc  := CM_Error; { default is sound the bell }

end. { CRTMENU }